//
//  main.c
//  SMCSMicroReceipt_Sample
//
//  Created by k on 2017/05/05.
//  Copyright © 2017年 MyCompany. All rights reserved.
//

#include <stdio.h>
#include <cups/cups.h>
#include "SMCSMicroReceipt.h"

void GetMicroReceiptTypeUsage();
void SetMicroReceiptTypeUsage(SMCSMicroReceiptType type);

char g_printerName[100];

int printTestPage(const char *printerName) {
    if (printerName == NULL) {
        fprintf(stderr, "Invalid printer name.\n");
        return -1;
    }
    
    int i;
    cups_dest_t *dests, *dest;
    int num_dests = cupsGetDests(&dests);
    
    int isSuccess = 0;
    
    for (i = num_dests, dest = dests; i > 0; i--, dest++) {
        // Get Printer
        const char *value = cupsGetOption("printer-info", dest->num_options, dest->options);
        if (value == NULL) {
            fprintf(stderr, "Failed to get option\n");
            return -2;
        }
        
        if (strlen(value) == 0) {
            value = dest->name;
        }
        
        if (strncmp(value, printerName, 100) != 0) {
            continue;
        }
        
        // Set Options
        int num_options = 0;
        cups_option_t *options = NULL;
        
        for (i = 0; i < dest->num_options; i++) {
            num_options = cupsAddOption(dest->options[i].name, dest->options[i].value,
                                        num_options, &options);
        }

        // Print
        int job_id = cupsPrintFile(dest->name, "/usr/share/cups/data/testprint",
                                   "Test Print", num_options, options);
        
        cupsFreeOptions(num_options, options);
        
        if (job_id == 0) {
            return -3;
        }
        
        isSuccess = 1;
        break;
    }
    
    cupsFreeDests(num_dests, dests);
    
    if (isSuccess == 0) {
        fprintf(stderr, "Printer not found.\n");
    }

    return 0;
}

int main(int argc, const char * argv[]) {
    printf("Enter the printer name: ");
    fgets(g_printerName, sizeof(g_printerName), stdin);
    
    g_printerName[strlen(g_printerName) - 1] = '\0';

    while (1) {
        char line[100] = { 0 };
        int selectedNum = -1;
        
        printf("[1] Get MicroReceipt\n");
        printf("[2] Set MicroReceipt(Disable) + Test Print\n");
        printf("[3] Set MicroReceipt(Information Logo + QR code)  + Test Print\n");
        printf("[4] Set MicroReceipt(Only QR code)  + Test Print\n");
        printf("[0] Quit\n");
        
        printf("Enter option number: ");
        fgets(line, sizeof(line), stdin);
        sscanf(line, "%d", &selectedNum);

        int finFlag = 0;
        switch (selectedNum) {
            case 1:
                GetMicroReceiptTypeUsage();
                break;
            case 2:
                SetMicroReceiptTypeUsage(DISABLE);
                break;
            case 3:
                SetMicroReceiptTypeUsage(INFORMATION_WITH_QRCODE);
                break;
            case 4:
                SetMicroReceiptTypeUsage(ONLY_QRCODE);
                break;
            case 0:
                finFlag = 1;
                break;
            default:
                fprintf(stderr, "Invalid Option.\n");
                break;
        }
        
        if (finFlag) {
            break;
        }
    }

    return 0;
}

void ShowFailureMessage(SMCSMicroReceiptResult result) {
    switch (result) {
        case SMCS_NOT_REGISTERED:
            fprintf(stderr, "Failure - SMCS_NOT_REGISTERED \n");
            break;
        case PRINTER_NAME_NOT_FOUND:
            fprintf(stderr, "Failure - PRINTER_NAME_NOT_FOUND \n");
            break;
        case MICRO_RECEIPT_PARAMETER_NOT_FOUND:
            fprintf(stderr, "Failure - MICRORECEIPT_PARMETER_NOT_FOUND \n");
            break;
        case SETTING_VALUE_IS_INCORRECT:
            fprintf(stderr, "Failure - SETTING_VALUE_IS_INCORRECT \n");
            break;
        case INITIALIZE_ERROR:
            fprintf(stderr, "Failure - INITIALIZE_ERROR \n");
            break;
        case SETTING_FILE_NOT_BE_LOADED:
            fprintf(stderr, "Failure - SETTING_FILE_NOT_BE_LOADED \n");
            break;
        case TIMEOUT:
            fprintf(stderr, "Failure - TIMEOUT \n");
            break;
        case MUTEX_ERROR:
            fprintf(stderr, "Failure - TIMEMUTEX_ERROROUT \n");
            break;
        default:
            fprintf(stderr, "Unexpected result: %d\n", result);
            break;
    }
}

void GetMicroReceiptTypeUsage() {
    SMCSMicroReceiptType type;
    
    SMCSMicroReceiptResult result = GetMicroReceiptType(g_printerName, &type);
    
    if (result == SUCCESS) {
        switch (type) {
            case DISABLE:
                printf("Success - Disable\n");
                break;
            case INFORMATION_WITH_QRCODE:
                printf("Success - Information Logo + QR code\n");
                break;
            case ONLY_QRCODE:
                printf("Success - QR code\n");
                break;
            default:
                fprintf(stderr, "Unexpected type: %d\n", type);
        }
    }
    else {
        ShowFailureMessage(result);
    }
}

void SetMicroReceiptTypeUsage(SMCSMicroReceiptType type) {

    SMCSMicroReceiptResult result = SetMicroReceiptType(g_printerName, type);
    if (result == SUCCESS) {
        printTestPage(g_printerName);
        
        printf("Success\n");
    }
    else {
        ShowFailureMessage(result);
    }
}

