//
//  SMCSMicroReceipt.h
//
//  Copyright © 2017 Star Micronics co.,ltd. All rights reserved.
//

typedef enum {
    DISABLE = 0,                            ///< Micro receipt setting is "disable".
    INFORMATION_WITH_QRCODE = 1,            ///< Micro receipt setting is "Information Logo + QR code".
    ONLY_QRCODE = 2                         ///< Micro receipt setting is "QR code".
} SMCSMicroReceiptType;

typedef enum {
    SUCCESS = 0,                            ///< Completed Successfully.
    SMCS_NOT_REGISTERED = 1,                ///< Star Cloud Services is unregistered.
    PRINTER_NAME_NOT_FOUND = 2,             ///< Cannot find the printer name specified. Please specify the printer queue name
    MICRO_RECEIPT_PARAMETER_NOT_FOUND = 3,  ///< Cannot find the micro receipt parameter
    SETTING_VALUE_IS_INCORRECT = 4,         ///< SMCSMicroReceiptType value setting is wrong. Please check the printer model.
    INITIALIZE_ERROR = 5,                   ///< Failed to initialize
    SETTING_FILE_NOT_BE_LOADED = 6,         ///< Failed to load the setting file.
    TIMEOUT = 7,                            ///< Operation timed out.
    MUTEX_ERROR = 8                         ///< Failed to get mutex
} SMCSMicroReceiptResult;

/**
 The SetMicroReceiptType function sets MicroReceipt value.
 @note   Please close the Utility(Star Cloud Services.app) and use this function.
 
 @param  printerName Please specify the device (printer queue) registered with the Star Cloud Services account.
                     Please refer to the software manual of each utility.
 @param  type        Pointer to a SMCSMicroReceiptType enum.
 @return SMCSMicroReceiptResult enum. If the function succeeds, the return value is SUCCESS. If the function failes, the return value is not-SUCCESS
 */
SMCSMicroReceiptResult GetMicroReceiptType(const char *printerName, SMCSMicroReceiptType *type);

/**
 The SetMicroReceiptType function sets Micro receipt value.
 @note   Please close the Utility(Star Cloud Services.app) and use this function.
 
 @param  printerName Please specify the device (printer queue) registered with the Star Cloud Services account.
                     Please refer to the software manual of each utility.
 @param  type        Specify SMCSMicroReceiptType enum. This value can be the following value.
                     (DISABLE, INFORMATION_WITH_QRCODE, ONLY_QRCODE)
 */
SMCSMicroReceiptResult SetMicroReceiptType(const char *printerName, SMCSMicroReceiptType type);
